@@ -38,6 +38,16 @@
 #define PMK_LEN 32
 #define EAP_PSK_LEN 16
 
+
+#define DEFAULT_EAP_WORKAROUND ((unsigned int) -1)
+#define DEFAULT_EAPOL_FLAGS (EAPOL_FLAG_REQUIRE_KEY_UNICAST | \
+			     EAPOL_FLAG_REQUIRE_KEY_BROADCAST)
+#define DEFAULT_PROTO (WPA_PROTO_WPA | WPA_PROTO_RSN)
+#define DEFAULT_KEY_MGMT (WPA_KEY_MGMT_PSK | WPA_KEY_MGMT_IEEE8021X)
+#define DEFAULT_PAIRWISE (WPA_CIPHER_CCMP | WPA_CIPHER_TKIP)
+#define DEFAULT_GROUP (WPA_CIPHER_CCMP | WPA_CIPHER_TKIP | \
+		       WPA_CIPHER_WEP104 | WPA_CIPHER_WEP40)
+
 /**
  * struct wpa_ssid - Network configuration data
  *
@@ -241,18 +251,43 @@ struct wpa_ssid {
 	 * ca_cert - File path to CA certificate file (PEM/DER)
 	 *
 	 * This file can have one or more trusted CA certificates. If ca_cert
-	 * is not included, server certificate will not be verified. This is
-	 * insecure and the CA file should always be configured when using
-	 * EAP-TLS/TTLS/PEAP.
+	 * and ca_path are not included, server certificate will not be
+	 * verified. This is insecure and a trusted CA certificate should
+	 * always be configured when using EAP-TLS/TTLS/PEAP. Full path to the
+	 * file should be used since working directory may change when
+	 * wpa_supplicant is run in the background.
+	 *
+	 * Alternatively, a named configuration blob can be used by setting
+	 * this to blob://<blob name>.
+	 *
+	 * On Windows, trusted CA certificates can be loaded from the system
+	 * certificate store by setting this to cert_store://<name>, e.g.,
+	 * ca_cert="cert_store://CA" or ca_cert="cert_store://ROOT".
 	 */
 	u8 *ca_cert;
 
+	/**
+	 * ca_path - Directory path for CA certificate files (PEM)
+	 *
+	 * This path may contain multiple CA certificates in OpenSSL format.
+	 * Common use for this is to point to system trusted CA list which is
+	 * often installed into directory like /etc/ssl/certs. If configured,
+	 * these certificates are added to the list of trusted CAs. ca_cert
+	 * may also be included in that case, but it is not required.
+	 */
+	u8 *ca_path;
+
 	/**
 	 * client_cert - File path to client certificate file (PEM/DER)
 	 *
 	 * This field is used with EAP method that use TLS authentication.
 	 * Usually, this is only configured for EAP-TLS, even though this could
-	 * in theory be used with EAP-TTLS and EAP-PEAP, too.
+	 * in theory be used with EAP-TTLS and EAP-PEAP, too. Full path to the
+	 * file should be used since working directory may change when
+	 * wpa_supplicant is run in the background.
+	 *
+	 * Alternatively, a named configuration blob can be used by setting
+	 * this to blob://<blob name>.
 	 */
 	u8 *client_cert;
 
@@ -261,7 +296,9 @@ struct wpa_ssid {
 	 *
 	 * When PKCS#12/PFX file (.p12/.pfx) is used, client_cert should be
 	 * commented out. Both the private key and certificate will be read
-	 * from the PKCS#12 file in this case.
+	 * from the PKCS#12 file in this case. Full path to the file should be
+	 * used since working directory may change when wpa_supplicant is run
+	 * in the background.
 	 *
 	 * Windows certificate store can be used by leaving client_cert out and
 	 * configuring private_key in one of the following formats:
@@ -271,6 +308,9 @@ struct wpa_ssid {
 	 * hash://certificate_thumbprint_in_hex
 	 *
 	 * For example: private_key="hash://63093aa9c47f56ae88334c7b65a4"
+	 *
+	 * Alternatively, a named configuration blob can be used by setting
+	 * this to blob://<blob name>.
 	 */
 	u8 *private_key;
 
@@ -290,7 +330,12 @@ struct wpa_ssid {
 	 * possible setup RSA to use ephemeral DH key exchange. In addition,
 	 * ciphers with DSA keys always use ephemeral DH keys. This can be used
 	 * to achieve forward secrecy. If the file is in DSA parameters format,
-	 * it will be automatically converted into DH params.
+	 * it will be automatically converted into DH params. Full path to the
+	 * file should be used since working directory may change when
+	 * wpa_supplicant is run in the background.
+	 *
+	 * Alternatively, a named configuration blob can be used by setting
+	 * this to blob://<blob name>.
 	 */
 	u8 *dh_file;
 
@@ -326,19 +371,44 @@ struct wpa_ssid {
 	 * ca_cert2 - File path to CA certificate file (PEM/DER) (Phase 2)
 	 *
 	 * This file can have one or more trusted CA certificates. If ca_cert2
-	 * is not included, server certificate will not be verified. This is
-	 * insecure and the CA file should always be configured.
+	 * and ca_path2 are not included, server certificate will not be
+	 * verified. This is insecure and a trusted CA certificate should
+	 * always be configured. Full path to the file should be used since
+	 * working directory may change when wpa_supplicant is run in the
+	 * background.
 	 *
 	 * This field is like ca_cert, but used for phase 2 (inside
 	 * EAP-TTLS/PEAP/FAST tunnel) authentication.
+	 *
+	 * Alternatively, a named configuration blob can be used by setting
+	 * this to blob://<blob name>.
 	 */
 	u8 *ca_cert2;
 
+	/**
+	 * ca_path2 - Directory path for CA certificate files (PEM) (Phase 2)
+	 *
+	 * This path may contain multiple CA certificates in OpenSSL format.
+	 * Common use for this is to point to system trusted CA list which is
+	 * often installed into directory like /etc/ssl/certs. If configured,
+	 * these certificates are added to the list of trusted CAs. ca_cert
+	 * may also be included in that case, but it is not required.
+	 *
+	 * This field is like ca_path, but used for phase 2 (inside
+	 * EAP-TTLS/PEAP/FAST tunnel) authentication.
+	 */
+	u8 *ca_path2;
+
 	/**
 	 * client_cert2 - File path to client certificate file
 	 *
 	 * This field is like client_cert, but used for phase 2 (inside
-	 * EAP-TTLS/PEAP/FAST tunnel) authentication.
+	 * EAP-TTLS/PEAP/FAST tunnel) authentication. Full path to the
+	 * file should be used since working directory may change when
+	 * wpa_supplicant is run in the background.
+	 *
+	 * Alternatively, a named configuration blob can be used by setting
+	 * this to blob://<blob name>.
 	 */
 	u8 *client_cert2;
 
@@ -346,7 +416,12 @@ struct wpa_ssid {
 	 * private_key2 - File path to client private key file
 	 *
 	 * This field is like private_key, but used for phase 2 (inside
-	 * EAP-TTLS/PEAP/FAST tunnel) authentication.
+	 * EAP-TTLS/PEAP/FAST tunnel) authentication. Full path to the
+	 * file should be used since working directory may change when
+	 * wpa_supplicant is run in the background.
+	 *
+	 * Alternatively, a named configuration blob can be used by setting
+	 * this to blob://<blob name>.
 	 */
 	u8 *private_key2;
 
@@ -362,7 +437,12 @@ struct wpa_ssid {
 	 * dh_file2 - File path to DH/DSA parameters file (in PEM format)
 	 *
 	 * This field is like dh_file, but used for phase 2 (inside
-	 * EAP-TTLS/PEAP/FAST tunnel) authentication.
+	 * EAP-TTLS/PEAP/FAST tunnel) authentication. Full path to the
+	 * file should be used since working directory may change when
+	 * wpa_supplicant is run in the background.
+	 *
+	 * Alternatively, a named configuration blob can be used by setting
+	 * this to blob://<blob name>.
 	 */
 	u8 *dh_file2;
 
@@ -624,10 +704,14 @@ struct wpa_ssid {
 	unsigned int eap_workaround;
 
 	/**
-	 * pac_file - File path for the PAC entries (EAP-FAST)
+	 * pac_file - File path or blob name for the PAC entries (EAP-FAST)
 	 *
 	 * wpa_supplicant will need to be able to create this file and write
-	 * updates to it when PAC is being provisioned or refreshed.
+	 * updates to it when PAC is being provisioned or refreshed. Full path
+	 * to the file should be used since working directory may change when
+	 * wpa_supplicant is run in the background.
+	 * Alternatively, a named configuration blob can be used by setting
+	 * this to blob://<blob name>.
 	 */
 	char *pac_file;
 
