@@ -1,6 +1,6 @@
 /*
  * WPA Supplicant - driver interaction with generic Linux Wireless Extensions
- * Copyright (c) 2003-2005, Jouni Malinen <jkmaline@cc.hut.fi>
+ * Copyright (c) 2003-2006, Jouni Malinen <jkmaline@cc.hut.fi>
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License version 2 as
@@ -42,6 +42,8 @@ struct wpa_driver_wext_data {
 	int event_sock;
 	int ioctl_sock;
 	char ifname[IFNAMSIZ + 1];
+	int ifindex;
+	int ifindex2;
 	u8 *assoc_req_ies;
 	size_t assoc_req_ies_len;
 	u8 *assoc_resp_ies;
@@ -57,20 +59,20 @@ static int wpa_driver_wext_get_range(void *priv);
 
 
 static int wpa_driver_wext_set_auth_param(struct wpa_driver_wext_data *drv,
-					  int index, u32 value)
+					  int idx, u32 value)
 {
 	struct iwreq iwr;
 	int ret = 0;
 
 	memset(&iwr, 0, sizeof(iwr));
 	strncpy(iwr.ifr_name, drv->ifname, IFNAMSIZ);
-	iwr.u.param.flags = index & IW_AUTH_INDEX;
+	iwr.u.param.flags = idx & IW_AUTH_INDEX;
 	iwr.u.param.value = value;
 
 	if (ioctl(drv->ioctl_sock, SIOCSIWAUTH, &iwr) < 0) {
-		fprintf(stderr, "WEXT auth param %d value 0x%x - ",
-			index, value);
 		perror("ioctl[SIOCSIWAUTH]");
+		fprintf(stderr, "WEXT auth param %d value 0x%x - ",
+			idx, value);
 		ret = -1;
 	}
 
@@ -538,7 +540,12 @@ static void wpa_driver_wext_event_rtm_newlink(struct wpa_driver_wext_data *drv,
 
 	ifi = NLMSG_DATA(h);
 
-	/* TODO: use ifi->ifi_index to recognize the interface (?) */
+	if (drv->ifindex != ifi->ifi_index && drv->ifindex2 != ifi->ifi_index)
+	{
+		wpa_printf(MSG_DEBUG, "Ignore event for foreign ifindex %d",
+			   ifi->ifi_index);
+		return;
+	}
 
 	nlmsg_len = NLMSG_ALIGN(sizeof(struct ifinfomsg));
 
@@ -758,6 +765,23 @@ void * wpa_driver_wext_init(void *ctx, const char *ifname)
 
 	wpa_driver_wext_get_range(drv);
 
+	drv->ifindex = if_nametoindex(drv->ifname);
+
+	if (strncmp(ifname, "wlan", 4) == 0) {
+		/*
+		 * Host AP driver may use both wlan# and wifi# interface in
+		 * wireless events. Since some of the versions included WE-18
+		 * support, let's add the alternative ifindex also from
+		 * driver_wext.c for the time being. This may be removed at
+		 * some point once it is believed that old versions of the
+		 * driver are not in use anymore.
+		 */
+		char ifname2[IFNAMSIZ + 1];
+		strncpy(ifname2, ifname, sizeof(ifname2));
+		memcpy(ifname2, "wifi", 4);
+		wpa_driver_wext_alternative_ifindex(drv, ifname2);
+	}
+
 	return drv;
 }
 
@@ -1012,7 +1036,7 @@ int wpa_driver_wext_get_scan_results(void *priv,
 			break;
 		case SIOCGIWFREQ:
 			if (ap_num < max_size) {
-				int div = 1000000, i;
+				int divi = 1000000, i;
 				if (iwe->u.freq.e > 6) {
 					wpa_printf(
 						MSG_DEBUG, "Invalid freq "
@@ -1023,8 +1047,8 @@ int wpa_driver_wext_get_scan_results(void *priv,
 					break;
 				}
 				for (i = 0; i < iwe->u.freq.e; i++)
-					div /= 10;
-				results[ap_num].freq = iwe->u.freq.m / div;
+					divi /= 10;
+				results[ap_num].freq = iwe->u.freq.m / divi;
 			}
 			break;
 		case IWEVQUAL:
@@ -1326,7 +1350,7 @@ static int wpa_driver_wext_set_key_ext(void *priv, wpa_alg alg,
  *	%WPA_ALG_TKIP, %WPA_ALG_CCMP); %WPA_ALG_NONE clears the key.
  * @addr: Address of the peer STA or ff:ff:ff:ff:ff:ff for
  *	broadcast/default keys
- * @key_idx: Key index (0..3), always 0 for unicast keys
+ * @key_idx: key index (0..3), usually 0 for unicast keys
  * @set_tx: Configure this key as the default Tx key (only used when
  *	driver does not support separate unicast/individual key
  * @seq: Sequence number/packet number, seq_len octets, the next
@@ -1662,7 +1686,8 @@ static int wpa_driver_wext_pmksa(struct wpa_driver_wext_data *drv,
 	iwr.u.data.length = sizeof(pmksa);
 
 	if (ioctl(drv->ioctl_sock, SIOCSIWPMKSA, &iwr) < 0) {
-		perror("ioctl[SIOCSIWPMKSA]");
+		if (errno != EOPNOTSUPP)
+			perror("ioctl[SIOCSIWPMKSA]");
 		ret = -1;
 	}
 
@@ -1703,6 +1728,25 @@ static int wpa_driver_wext_get_capa(void *priv, struct wpa_driver_capa *capa)
 }
 
 
+int wpa_driver_wext_alternative_ifindex(struct wpa_driver_wext_data *drv,
+					const char *ifname)
+{
+	if (ifname == NULL) {
+		drv->ifindex2 = -1;
+		return 0;
+	}
+
+	drv->ifindex2 = if_nametoindex(ifname);
+	if (drv->ifindex2 <= 0)
+		return -1;
+
+	wpa_printf(MSG_DEBUG, "Added alternative ifindex %d (%s) for "
+		   "wireless events", drv->ifindex2, ifname);
+
+	return 0;
+}
+
+
 const struct wpa_driver_ops wpa_driver_wext_ops = {
 	.name = "wext",
 	.desc = "Linux wireless extensions (generic)",
