@@ -1,6 +1,6 @@
 /*
  * WPA Supplicant
- * Copyright (c) 2003-2005, Jouni Malinen <jkmaline@cc.hut.fi>
+ * Copyright (c) 2003-2006, Jouni Malinen <jkmaline@cc.hut.fi>
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License version 2 as
@@ -22,7 +22,6 @@
 #ifndef CONFIG_NATIVE_WINDOWS
 #include <netinet/in.h>
 #endif /* CONFIG_NATIVE_WINDOWS */
-#include <fcntl.h>
 
 #include "common.h"
 #include "eapol_sm.h"
@@ -39,11 +38,11 @@
 #include "preauth.h"
 #include "wpa_ctrl.h"
 
-static const char *wpa_supplicant_version =
+const char *wpa_supplicant_version =
 "wpa_supplicant v" VERSION_STR "\n"
-"Copyright (c) 2003-2005, Jouni Malinen <jkmaline@cc.hut.fi> and contributors";
+"Copyright (c) 2003-2006, Jouni Malinen <jkmaline@cc.hut.fi> and contributors";
 
-static const char *wpa_supplicant_license =
+const char *wpa_supplicant_license =
 "This program is free software. You can distribute it and/or modify it\n"
 "under the terms of the GNU General Public License version 2.\n"
 "\n"
@@ -56,7 +55,7 @@ static const char *wpa_supplicant_license =
 ;
 
 #ifndef CONFIG_NO_STDOUT_DEBUG
-static const char *wpa_supplicant_full_license =
+const char *wpa_supplicant_full_license =
 "This program is free software; you can redistribute it and/or modify\n"
 "it under the terms of the GNU General Public License version 2 as\n"
 "published by the Free Software Foundation.\n"
@@ -135,9 +134,9 @@ void wpa_msg(struct wpa_supplicant *wpa_s, int level, char *fmt, ...)
 }
 
 
-u8 * wpa_alloc_eapol(const struct wpa_supplicant *wpa_s, u8 type,
-		     const void *data, u16 data_len,
-		     size_t *msg_len, void **data_pos)
+static u8 * wpa_alloc_eapol(const struct wpa_supplicant *wpa_s, u8 type,
+			    const void *data, u16 data_len,
+			    size_t *msg_len, void **data_pos)
 {
 	struct ieee802_1x_hdr *hdr;
 
@@ -192,7 +191,8 @@ int wpa_ether_send(struct wpa_supplicant *wpa_s, const u8 *dest, u16 proto,
  * This function adds Ethernet and IEEE 802.1X header and sends the EAPOL frame
  * to the current Authenticator.
  */
-static int wpa_supplicant_eapol_send(void *ctx, int type, u8 *buf, size_t len)
+static int wpa_supplicant_eapol_send(void *ctx, int type, const u8 *buf,
+				     size_t len)
 {
 	struct wpa_supplicant *wpa_s = ctx;
 	u8 *msg, *dst, bssid[ETH_ALEN];
@@ -266,7 +266,7 @@ static int wpa_supplicant_eapol_send(void *ctx, int type, u8 *buf, size_t len)
  * Returns 0 on success or < 0 on error.
  */
 static int wpa_eapol_set_wep_key(void *ctx, int unicast, int keyidx,
-				 u8 *key, size_t keylen)
+				 const u8 *key, size_t keylen)
 {
 	struct wpa_supplicant *wpa_s = ctx;
 	return wpa_drv_set_key(wpa_s, WPA_ALG_WEP,
@@ -277,6 +277,24 @@ static int wpa_eapol_set_wep_key(void *ctx, int unicast, int keyidx,
 #endif /* IEEE8021X_EAPOL */
 
 
+#if defined(IEEE8021X_EAPOL) || !defined(CONFIG_NO_WPA)
+static void wpa_supplicant_set_config_blob(void *ctx,
+					   struct wpa_config_blob *blob)
+{
+	struct wpa_supplicant *wpa_s = ctx;
+	wpa_config_set_blob(wpa_s->conf, blob);
+}
+
+
+static const struct wpa_config_blob *
+wpa_supplicant_get_config_blob(void *ctx, const char *name)
+{
+	struct wpa_supplicant *wpa_s = ctx;
+	return wpa_config_get_blob(wpa_s->conf, name);
+}
+#endif /* defined(IEEE8021X_EAPOL) || !defined(CONFIG_NO_WPA) */
+
+
 /* Configure default/group WEP key for static WEP */
 static int wpa_set_wep_key(void *ctx, int set_tx, int keyidx, const u8 *key,
 			   size_t keylen)
@@ -539,6 +557,8 @@ void wpa_supplicant_initiate_eapol(struct wpa_supplicant *wpa_s)
 	}
 	eapol_conf.fast_reauth = wpa_s->conf->fast_reauth;
 	eapol_conf.workaround = ssid->eap_workaround;
+	eapol_conf.eap_disabled = wpa_s->key_mgmt != WPA_KEY_MGMT_IEEE8021X &&
+		wpa_s->key_mgmt != WPA_KEY_MGMT_IEEE8021X_NO_WPA;
 	eapol_sm_notify_config(wpa_s->eapol, ssid, &eapol_conf);
 }
 
@@ -612,6 +632,8 @@ static void wpa_supplicant_cleanup(struct wpa_supplicant *wpa_s)
 	free(wpa_s->scan_results);
 	wpa_s->scan_results = NULL;
 	wpa_s->num_scan_results = 0;
+
+	wpa_supplicant_cancel_scan(wpa_s);
 }
 
 
@@ -698,8 +720,9 @@ wpa_states wpa_supplicant_get_state(struct wpa_supplicant *wpa_s)
 static void wpa_supplicant_terminate(int sig, void *eloop_ctx,
 				     void *signal_ctx)
 {
-	struct wpa_supplicant *wpa_s = eloop_ctx;
-	for (wpa_s = wpa_s->head; wpa_s; wpa_s = wpa_s->next) {
+	struct wpa_global *global = eloop_ctx;
+	struct wpa_supplicant *wpa_s;
+	for (wpa_s = global->ifaces; wpa_s; wpa_s = wpa_s->next) {
 		wpa_msg(wpa_s, MSG_INFO, WPA_EVENT_TERMINATING "- signal %d "
 			"received", sig);
 	}
@@ -753,9 +776,10 @@ int wpa_supplicant_reload_configuration(struct wpa_supplicant *wpa_s)
 static void wpa_supplicant_reconfig(int sig, void *eloop_ctx,
 				    void *signal_ctx)
 {
-	struct wpa_supplicant *wpa_s = eloop_ctx;
+	struct wpa_global *global = eloop_ctx;
+	struct wpa_supplicant *wpa_s;
 	wpa_printf(MSG_DEBUG, "Signal %d received - reconfiguring", sig);
-	for (wpa_s = wpa_s->head; wpa_s; wpa_s = wpa_s->next) {
+	for (wpa_s = global->ifaces; wpa_s; wpa_s = wpa_s->next) {
 		if (wpa_supplicant_reload_configuration(wpa_s) < 0) {
 			eloop_terminate();
 		}
@@ -908,10 +932,14 @@ static wpa_key_mgmt key_mgmt2driver(int key_mgmt)
 
 static int wpa_supplicant_suites_from_ai(struct wpa_supplicant *wpa_s,
 					 struct wpa_ssid *ssid,
-					 struct wpa_ie_data *ie) {
-	if (wpa_sm_parse_own_wpa_ie(wpa_s->wpa, ie)) {
-		wpa_msg(wpa_s, MSG_INFO, "WPA: Failed to parse WPA IE from "
-			"association info");
+					 struct wpa_ie_data *ie)
+{
+	int ret = wpa_sm_parse_own_wpa_ie(wpa_s->wpa, ie);
+	if (ret) {
+		if (ret == -2) {
+			wpa_msg(wpa_s, MSG_INFO, "WPA: Failed to parse WPA IE "
+				"from association info");
+		}
 		return -1;
 	}
 
@@ -947,26 +975,29 @@ int wpa_supplicant_set_suites(struct wpa_supplicant *wpa_s,
 {
 	struct wpa_ie_data ie;
 	int sel, proto;
-	u8 *ap_ie;
-	size_t ap_ie_len;
 
-	if (bss && bss->rsn_ie_len && (ssid->proto & WPA_PROTO_RSN)) {
+	if (bss && bss->rsn_ie_len && (ssid->proto & WPA_PROTO_RSN) &&
+	    wpa_parse_wpa_ie(bss->rsn_ie, bss->rsn_ie_len, &ie) == 0 &&
+	    (ie.group_cipher & ssid->group_cipher) &&
+	    (ie.pairwise_cipher & ssid->pairwise_cipher) &&
+	    (ie.key_mgmt & ssid->key_mgmt)) {
 		wpa_msg(wpa_s, MSG_DEBUG, "RSN: using IEEE 802.11i/D9.0");
 		proto = WPA_PROTO_RSN;
-		ap_ie = bss->rsn_ie;
-		ap_ie_len = bss->rsn_ie_len;
-	} else if (bss) {
+	} else if (bss && bss->wpa_ie_len && (ssid->proto & WPA_PROTO_WPA) &&
+		   wpa_parse_wpa_ie(bss->wpa_ie, bss->wpa_ie_len, &ie) == 0 &&
+		   (ie.group_cipher & ssid->group_cipher) &&
+		   (ie.pairwise_cipher & ssid->pairwise_cipher) &&
+		   (ie.key_mgmt & ssid->key_mgmt)) {
 		wpa_msg(wpa_s, MSG_DEBUG, "WPA: using IEEE 802.11i/D3.0");
 		proto = WPA_PROTO_WPA;
-		ap_ie = bss->wpa_ie;
-		ap_ie_len = bss->wpa_ie_len;
+	} else if (bss) {
+		wpa_msg(wpa_s, MSG_WARNING, "WPA: Failed to select WPA/RSN");
+		return -1;
 	} else {
 		if (ssid->proto & WPA_PROTO_RSN)
 			proto = WPA_PROTO_RSN;
 		else
 			proto = WPA_PROTO_WPA;
-		ap_ie = NULL;
-		ap_ie_len = 0;
 		if (wpa_supplicant_suites_from_ai(wpa_s, ssid, &ie) < 0) {
 			memset(&ie, 0, sizeof(ie));
 			ie.group_cipher = ssid->group_cipher;
@@ -977,11 +1008,6 @@ int wpa_supplicant_set_suites(struct wpa_supplicant *wpa_s,
 		}
 	}
 
-	if (ap_ie && wpa_parse_wpa_ie(ap_ie, ap_ie_len, &ie)) {
-		wpa_msg(wpa_s, MSG_WARNING, "WPA: Failed to parse WPA IE for "
-			"the selected BSS.");
-		return -1;
-	}
 	wpa_printf(MSG_DEBUG, "WPA: Selected cipher suites: group %d "
 		   "pairwise %d key_mgmt %d",
 		   ie.group_cipher, ie.pairwise_cipher, ie.key_mgmt);
@@ -1075,6 +1101,7 @@ void wpa_supplicant_associate(struct wpa_supplicant *wpa_s,
 	struct wpa_driver_associate_params params;
 	int wep_keys_set = 0;
 	struct wpa_driver_capa capa;
+	int assoc_failed = 0;
 
 	wpa_s->reassociate = 0;
 	if (bss) {
@@ -1210,6 +1237,7 @@ void wpa_supplicant_associate(struct wpa_supplicant *wpa_s,
 			"failed");
 		/* try to continue anyway; new association will be tried again
 		 * after timeout */
+		assoc_failed = 1;
 	}
 
 	if (wpa_s->key_mgmt == WPA_KEY_MGMT_WPA_NONE) {
@@ -1222,7 +1250,14 @@ void wpa_supplicant_associate(struct wpa_supplicant *wpa_s,
 		wpa_supplicant_set_state(wpa_s, WPA_COMPLETED);
 	} else {
 		/* Timeout for IEEE 802.11 authentication and association */
-		wpa_supplicant_req_auth_timeout(wpa_s, 5, 0);
+		int timeout;
+		if (assoc_failed)
+			timeout = 5;
+		else if (wpa_s->conf->ap_scan == 1)
+			timeout = 10;
+		else
+			timeout = 60;
+		wpa_supplicant_req_auth_timeout(wpa_s, timeout, 0);
 	}
 
 	if (wep_keys_set && wpa_drv_get_capa(wpa_s, &capa) == 0 &&
@@ -1531,24 +1566,6 @@ static int wpa_supplicant_set_driver(struct wpa_supplicant *wpa_s,
 }
 
 
-static void wpa_supplicant_fd_workaround(void)
-{
-	int s, i;
-	/* When started from pcmcia-cs scripts, wpa_supplicant might start with
-	 * fd 0, 1, and 2 closed. This will cause some issues because many
-	 * places in wpa_supplicant are still printing out to stdout. As a
-	 * workaround, make sure that fd's 0, 1, and 2 are not used for other
-	 * sockets. */
-	for (i = 0; i < 3; i++) {
-		s = open("/dev/null", O_RDWR);
-		if (s > 2) {
-			close(s);
-			break;
-		}
-	}
-}
-
-
 void wpa_supplicant_rx_eapol(void *ctx, const u8 *src_addr,
 			     const u8 *buf, size_t len)
 {
@@ -1659,48 +1676,23 @@ int wpa_supplicant_driver_init(struct wpa_supplicant *wpa_s,
 }
 
 
-static void usage(void)
+static int wpa_supplicant_daemon(const char *pid_file)
 {
-	int i;
-	printf("%s\n\n%s\n"
-	       "usage:\n"
-	       "  wpa_supplicant [-BddehLqqvwW] -i<ifname> -c<config file> "
-	       "[-D<driver>] \\\n"
-	       "      [-P<pid file>] "
-	       "[-N -i<ifname> -c<conf> [-D<driver>] ...]\n"
-	       "\n"
-	       "drivers:\n",
-	       wpa_supplicant_version, wpa_supplicant_license);
-
-	for (i = 0; wpa_supplicant_drivers[i]; i++) {
-		printf("  %s = %s\n",
-		       wpa_supplicant_drivers[i]->name,
-		       wpa_supplicant_drivers[i]->desc);
+	wpa_printf(MSG_DEBUG, "Daemonize..");
+	if (daemon(0, 0)) {
+		perror("daemon");
+		return -1;
 	}
 
-#ifndef CONFIG_NO_STDOUT_DEBUG
-	printf("options:\n"
-	       "  -B = run daemon in the background\n"
-	       "  -d = increase debugging verbosity (-dd even more)\n"
-	       "  -K = include keys (passwords, etc.) in debug output\n"
-	       "  -t = include timestamp in debug messages\n"
-	       "  -h = show this help text\n"
-	       "  -L = show license (GPL and BSD)\n"
-	       "  -q = decrease debugging verbosity (-qq even less)\n"
-	       "  -v = show version\n"
-	       "  -w = wait for interface to be added, if needed\n"
-	       "  -W = wait for a control interface monitor before starting\n"
-	       "  -N = start describing new interface\n");
-#endif /* CONFIG_NO_STDOUT_DEBUG */
-}
+	if (pid_file) {
+		FILE *f = fopen(pid_file, "w");
+		if (f) {
+			fprintf(f, "%u\n", getpid());
+			fclose(f);
+		}
+	}
 
-
-static void license(void)
-{
-#ifndef CONFIG_NO_STDOUT_DEBUG
-	printf("%s\n\n%s\n",
-	       wpa_supplicant_version, wpa_supplicant_full_license);
-#endif /* CONFIG_NO_STDOUT_DEBUG */
+	return 0;
 }
 
 
@@ -1719,23 +1711,21 @@ static struct wpa_supplicant * wpa_supplicant_alloc(void)
 }
 
 
-struct wpa_interface {
-	const char *confname, *driver, *ifname;
-};
-
-
-static int wpa_supplicant_init(struct wpa_supplicant *wpa_s,
-			       struct wpa_interface *iface)
+static int wpa_supplicant_init_iface(struct wpa_supplicant *wpa_s,
+				     struct wpa_interface *iface)
 {
 	wpa_printf(MSG_DEBUG, "Initializing interface '%s' conf '%s' driver "
-		   "'%s'", iface->ifname, iface->confname,
-		   iface->driver ? iface->driver : "default");
+		   "'%s' ctrl_interface '%s'", iface->ifname,
+		   iface->confname ? iface->confname : "N/A",
+		   iface->driver ? iface->driver : "default",
+		   iface->ctrl_interface ? iface->ctrl_interface : "N/A");
 
 	if (wpa_supplicant_set_driver(wpa_s, iface->driver) < 0) {
 		return -1;
 	}
 
 	if (iface->confname) {
+#ifdef CONFIG_BACKEND_FILE
 		wpa_s->confname = rel2abs_path(iface->confname);
 		if (wpa_s->confname == NULL) {
 			wpa_printf(MSG_ERROR, "Failed to get absolute path "
@@ -1745,22 +1735,41 @@ static int wpa_supplicant_init(struct wpa_supplicant *wpa_s,
 		}
 		wpa_printf(MSG_DEBUG, "Configuration file '%s' -> '%s'",
 			   iface->confname, wpa_s->confname);
+#else /* CONFIG_BACKEND_FILE */
+		wpa_s->confname = strdup(iface->confname);
+#endif /* CONFIG_BACKEND_FILE */
 		wpa_s->conf = wpa_config_read(wpa_s->confname);
 		if (wpa_s->conf == NULL) {
-			printf("Failed to read configuration file '%s'.\n",
-			       wpa_s->confname);
+			printf("Failed to read read or parse configuration "
+			       "'%s'.\n", wpa_s->confname);
 			return -1;
 		}
-	}
+
+		/*
+		 * Override ctrl_interface and driver_param if set on command
+		 * line.
+		 */
+		if (iface->ctrl_interface) {
+			free(wpa_s->conf->ctrl_interface);
+			wpa_s->conf->ctrl_interface =
+				strdup(iface->ctrl_interface);
+		}
+
+		if (iface->driver_param) {
+			free(wpa_s->conf->driver_param);
+			wpa_s->conf->driver_param =
+				strdup(iface->driver_param);
+		}
+	} else
+		wpa_s->conf = wpa_config_alloc_empty(iface->ctrl_interface,
+						     iface->driver_param);
 
 	if (wpa_s->conf == NULL) {
-		usage();
 		printf("\nNo configuration found.\n");
 		return -1;
 	}
 
 	if (iface->ifname == NULL) {
-		usage();
 		printf("\nInterface name is required.\n");
 		return -1;
 	}
@@ -1792,6 +1801,8 @@ static int wpa_supplicant_init_eapol(struct wpa_supplicant *wpa_s)
 	ctx->eapol_done_cb = wpa_supplicant_notify_eapol_done;
 	ctx->eapol_send = wpa_supplicant_eapol_send;
 	ctx->set_wep_key = wpa_eapol_set_wep_key;
+	ctx->set_config_blob = wpa_supplicant_set_config_blob;
+	ctx->get_config_blob = wpa_supplicant_get_config_blob;
 	ctx->opensc_engine_path = wpa_s->conf->opensc_engine_path;
 	ctx->pkcs11_engine_path = wpa_s->conf->pkcs11_engine_path;
 	ctx->pkcs11_module_path = wpa_s->conf->pkcs11_module_path;
@@ -1834,7 +1845,9 @@ static int wpa_supplicant_init_wpa(struct wpa_supplicant *wpa_s)
 	ctx->cancel_auth_timeout = _wpa_supplicant_cancel_auth_timeout;
 	ctx->add_pmkid = wpa_supplicant_add_pmkid;
 	ctx->remove_pmkid = wpa_supplicant_remove_pmkid;
-	
+	ctx->set_config_blob = wpa_supplicant_set_config_blob;
+	ctx->get_config_blob = wpa_supplicant_get_config_blob;
+
 	wpa_s->wpa = wpa_sm_init(ctx);
 	if (wpa_s->wpa == NULL) {
 		fprintf(stderr, "Failed to initialize WPA state machine\n");
@@ -1846,8 +1859,8 @@ static int wpa_supplicant_init_wpa(struct wpa_supplicant *wpa_s)
 }
 
 
-static int wpa_supplicant_init2(struct wpa_supplicant *wpa_s,
-				int wait_for_interface)
+static int wpa_supplicant_init_iface2(struct wpa_supplicant *wpa_s,
+				      int wait_for_interface)
 {
 	const char *ifname;
 
@@ -1935,9 +1948,11 @@ static int wpa_supplicant_init2(struct wpa_supplicant *wpa_s,
 }
 
 
-static void wpa_supplicant_deinit(struct wpa_supplicant *wpa_s)
+static void wpa_supplicant_deinit_iface(struct wpa_supplicant *wpa_s)
 {
 	if (wpa_s->drv_priv) {
+		wpa_supplicant_deauthenticate(wpa_s, REASON_DEAUTH_LEAVING);
+
 		/* Backwards compatibility call to set_wpa() handler. This is
 		 * called only just after init and just before deinit, so these
 		 * handler can be used to implement same functionality. */
@@ -1956,75 +1971,176 @@ static void wpa_supplicant_deinit(struct wpa_supplicant *wpa_s)
 }
 
 
-struct wpa_params {
-	int daemonize, wait_for_interface;
-	int wait_for_monitor;
-	char *pid_file;
-};
-
-static int wpa_supplicant_run(struct wpa_params *params,
-			      struct wpa_interface *iface, size_t num)
+/**
+ * wpa_supplicant_add_iface - Add a new network interface
+ * @global: Pointer to global data from wpa_supplicant_init()
+ * @iface: Interface configuration options
+ * Returns: Pointer to the created interface or %NULL on failure
+ *
+ * This function is used to add new network interfaces for %wpa_supplicant.
+ * This can be called before wpa_supplicant_run() to add interfaces before the
+ * main event loop has been started. In addition, new interfaces can be added
+ * dynamically while %wpa_supplicant is already running. This could happen,
+ * e.g., when a hotplug network adapter is inserted.
+ */
+struct wpa_supplicant * wpa_supplicant_add_iface(struct wpa_global *global,
+						 struct wpa_interface *iface)
 {
-	struct wpa_supplicant *head = NULL, *wpa_s = NULL, *new_wpa_s;
-	int i, exitcode;
+	struct wpa_supplicant *wpa_s;
 
-	for (i = 0; i < num; i++) {
-		new_wpa_s = wpa_supplicant_alloc();
-		if (new_wpa_s == NULL)
-			return -1;
-
-		if (head == NULL) {
-			head = wpa_s = new_wpa_s;
-			eloop_init(head);
-		} else {
-			wpa_s->next = new_wpa_s;
-			wpa_s = new_wpa_s;
-			wpa_s->head = head;
-		}
-
-		if (wpa_supplicant_init(wpa_s, &iface[i]))
+	if (global == NULL || iface == NULL)
+		return NULL;
+
+	wpa_s = wpa_supplicant_alloc();
+	if (wpa_s == NULL)
+		return NULL;
+
+	if (wpa_supplicant_init_iface(wpa_s, iface) ||
+	    wpa_supplicant_init_iface2(wpa_s,
+				       global->params.wait_for_interface)) {
+		wpa_printf(MSG_DEBUG, "Failed to add interface %s",
+			   iface->ifname);
+		wpa_supplicant_deinit_iface(wpa_s);
+		free(wpa_s);
+		return NULL;
+	}
+
+	wpa_s->global = global;
+	wpa_s->next = global->ifaces;
+	global->ifaces = wpa_s;
+
+	wpa_printf(MSG_DEBUG, "Added interface %s", wpa_s->ifname);
+
+	return wpa_s;
+}
+
+
+/**
+ * wpa_supplicant_remove_iface - Remove a network interface
+ * @global: Pointer to global data from wpa_supplicant_init()
+ * @wpa_s: Pointer to the network interface to be removed
+ * Returns: 0 if interface was removed, -1 if interface was not found
+ *
+ * This function can be used to dynamically remove network interfaces from
+ * %wpa_supplicant, e.g., when a hotplug network adapter is ejected. In
+ * addition, this function is used to remove all remaining interdaces when
+ * %wpa_supplicant is terminated.
+ */
+int wpa_supplicant_remove_iface(struct wpa_global *global,
+				struct wpa_supplicant *wpa_s)
+{
+	struct wpa_supplicant *prev;
+
+	/* Remove interface from the global list of interfaces */
+	prev = global->ifaces;
+	if (prev == wpa_s) {
+		global->ifaces = wpa_s->next;
+	} else {
+		while (prev && prev->next != wpa_s)
+			prev = prev->next;
+		if (prev == NULL)
 			return -1;
+		prev->next = wpa_s->next;
 	}
 
-	exitcode = 0;
+	wpa_printf(MSG_DEBUG, "Removing interface %s", wpa_s->ifname);
 
-	if (params->wait_for_interface && params->daemonize) {
-		wpa_printf(MSG_DEBUG, "Daemonize..");
-		if (daemon(0, 0)) {
-			perror("daemon");
-			exitcode = -1;
-			goto cleanup;
-		}
+	wpa_supplicant_deinit_iface(wpa_s);
+	free(wpa_s);
+
+	return 0;
+}
+
+
+/**
+ * wpa_supplicant_get_iface - Get a new network interface
+ * @global: Pointer to global data from wpa_supplicant_init()
+ * @ifname: Interface name
+ * Returns: Pointer to the interface or %NULL if not found
+ */
+struct wpa_supplicant * wpa_supplicant_get_iface(struct wpa_global *global,
+						 const char *ifname)
+{
+	struct wpa_supplicant *wpa_s;
+
+	for (wpa_s = global->ifaces; wpa_s; wpa_s = wpa_s->next) {
+		if (strcmp(wpa_s->ifname, ifname) == 0)
+			return wpa_s;
+	}
+	return NULL;
+}
+
+
+/**
+ * wpa_supplicant_init - Initialize %wpa_supplicant
+ * @params: Parameters for %wpa_supplicant
+ * Returns: Pointer to global %wpa_supplicant data, or %NULL on failure
+ *
+ * This function is used to initialize %wpa_supplicant. After successful
+ * initialization, the returned data pointer can be used to add and remove
+ * network interfaces, and eventually, to deinitialize %wpa_supplicant.
+ */
+struct wpa_global * wpa_supplicant_init(struct wpa_params *params)
+{
+	struct wpa_global *global;
+
+	if (params == NULL)
+		return NULL;
+	global = malloc(sizeof(*global));
+	if (global == NULL)
+		return NULL;
+	memset(global, 0, sizeof(*global));
+	global->params.daemonize = params->daemonize;
+	global->params.wait_for_interface = params->wait_for_interface;
+	global->params.wait_for_monitor = params->wait_for_monitor;
+	if (params->pid_file)
+		global->params.pid_file = strdup(params->pid_file);
+	if (params->ctrl_interface)
+		global->params.ctrl_interface = strdup(params->ctrl_interface);
+	wpa_debug_level = global->params.wpa_debug_level =
+		params->wpa_debug_level;
+	wpa_debug_show_keys = global->params.wpa_debug_show_keys =
+		params->wpa_debug_show_keys;
+	wpa_debug_timestamp = global->params.wpa_debug_timestamp =
+		params->wpa_debug_timestamp;
+
+	eloop_init(global);
+
+	if (wpa_supplicant_global_ctrl_iface_init(global)) {
+		eloop_destroy();
+		return NULL;
 	}
 
-	for (wpa_s = head; wpa_s; wpa_s = wpa_s->next) {
-		if (wpa_supplicant_init2(wpa_s, params->wait_for_interface)) {
-			exitcode = -1;
-			goto cleanup;
-		}
+	if (global->params.wait_for_interface && global->params.daemonize &&
+	    wpa_supplicant_daemon(global->params.pid_file)) {
+		eloop_destroy();
+		return NULL;
 	}
 
-	if (!params->wait_for_interface && params->daemonize) {
-		wpa_printf(MSG_DEBUG, "Daemonize..");
-		if (daemon(0, 0)) {
-			perror("daemon");
-			exitcode = -1;
-			goto cleanup;
-		}
-	}
+	return global;
+}
 
-	if (params->pid_file) {
-		FILE *f = fopen(params->pid_file, "w");
-		if (f) {
-			fprintf(f, "%u\n", getpid());
-			fclose(f);
-		}
-	}
 
-	if (params->wait_for_monitor) {
-		for (wpa_s = head; wpa_s; wpa_s = wpa_s->next) {
+/**
+ * wpa_supplicant_run - Run the %wpa_supplicant main event loop
+ * @global: Pointer to global data from wpa_supplicant_init()
+ * Returns: 0 after successful event loop run, -1 on failure
+ *
+ * This function starts the main event loop and continues running as long as
+ * there are any remaining events. In most cases, this function is running as
+ * long as the %wpa_supplicant process in still in use.
+ */
+int wpa_supplicant_run(struct wpa_global *global)
+{
+	struct wpa_supplicant *wpa_s;
+
+	if (!global->params.wait_for_interface && global->params.daemonize &&
+	    wpa_supplicant_daemon(global->params.pid_file))
+		return -1;
+
+	if (global->params.wait_for_monitor) {
+		for (wpa_s = global->ifaces; wpa_s; wpa_s = wpa_s->next)
 			wpa_supplicant_ctrl_iface_wait(wpa_s);
-		}
 	}
 
 	eloop_register_signal(SIGINT, wpa_supplicant_terminate, NULL);
@@ -2035,125 +2151,34 @@ static int wpa_supplicant_run(struct wpa_params *params,
 
 	eloop_run();
 
-	for (wpa_s = head; wpa_s; wpa_s = wpa_s->next) {
-		wpa_supplicant_deauthenticate(wpa_s, REASON_DEAUTH_LEAVING);
-	}
+	return 0;
+}
 
-cleanup:
-	wpa_s = head;
-	while (wpa_s) {
-		struct wpa_supplicant *prev;
-		wpa_supplicant_deinit(wpa_s);
-		prev = wpa_s;
-		wpa_s = wpa_s->next;
-		free(prev);
-	}
+
+/**
+ * wpa_supplicant_deinit - Deinitialize %wpa_supplicant
+ * @global: Pointer to global data from wpa_supplicant_init()
+ *
+ * This function is called to deinitialize %wpa_supplicant and to free all
+ * allocated resources. Remaining network interfaces will also be removed.
+ */
+void wpa_supplicant_deinit(struct wpa_global *global)
+{
+	if (global == NULL)
+		return;
+
+	while (global->ifaces)
+		wpa_supplicant_remove_iface(global, global->ifaces);
+
+	wpa_supplicant_global_ctrl_iface_deinit(global);
 
 	eloop_destroy();
 
-	if (params->pid_file)
-		unlink(params->pid_file);
-
-	return exitcode;
-}
-
-
-int main(int argc, char *argv[])
-{
-	int c;
-	struct wpa_interface *ifaces, *iface;
-	int iface_count, exitcode;
-	struct wpa_params params;
-
-#ifdef CONFIG_NATIVE_WINDOWS
-	WSADATA wsaData;
-	if (WSAStartup(MAKEWORD(2, 0), &wsaData)) {
-		printf("Could not find a usable WinSock.dll\n");
-		return -1;
+	if (global->params.pid_file) {
+		unlink(global->params.pid_file);
+		free(global->params.pid_file);
 	}
-#endif /* CONFIG_NATIVE_WINDOWS */
+	free(global->params.ctrl_interface);
 
-	memset(&params, 0, sizeof(params));
-	iface = ifaces = malloc(sizeof(struct wpa_interface));
-	if (ifaces == NULL)
-		return -1;
-	memset(iface, 0, sizeof(*iface));
-	iface_count = 1;
-
-	wpa_supplicant_fd_workaround();
-
-	for (;;) {
-		c = getopt(argc, argv, "Bc:D:dhi:KLNP:qtvwW");
-		if (c < 0)
-			break;
-		switch (c) {
-		case 'B':
-			params.daemonize++;
-			break;
-		case 'c':
-			iface->confname = optarg;
-			break;
-		case 'D':
-			iface->driver = optarg;
-			break;
-		case 'd':
-			wpa_debug_level--;
-			break;
-		case 'h':
-			usage();
-			return -1;
-		case 'i':
-			iface->ifname = optarg;
-			break;
-		case 'K':
-			wpa_debug_show_keys++;
-			break;
-		case 'L':
-			license();
-			return -1;
-		case 'P':
-			params.pid_file = rel2abs_path(optarg);
-			break;
-		case 'q':
-			wpa_debug_level++;
-			break;
-		case 't':
-			wpa_debug_timestamp++;
-			break;
-		case 'v':
-			printf("%s\n", wpa_supplicant_version);
-			return -1;
-		case 'w':
-			params.wait_for_interface++;
-			break;
-		case 'W':
-			params.wait_for_monitor++;
-			break;
-		case 'N':
-			iface_count++;
-			iface = realloc(ifaces, iface_count *
-					sizeof(struct wpa_interface));
-			if (ifaces == NULL) {
-				free(ifaces);
-				return -1;
-			}
-			ifaces = iface;
-			iface = &ifaces[iface_count - 1]; 
-			memset(iface, 0, sizeof(*iface));
-			break;
-		default:
-			usage();
-			return -1;
-		}
-	}
-
-	exitcode = wpa_supplicant_run(&params, ifaces, iface_count);
-	free(ifaces);
-	free(params.pid_file);
-
-#ifdef CONFIG_NATIVE_WINDOWS
-	WSACleanup();
-#endif /* CONFIG_NATIVE_WINDOWS */
-
-	return exitcode;
+	free(global);
 }
